package be.iminds.ilabt.jfed.highlevel.jobs.states;

import be.iminds.ilabt.jfed.espec.ESpecLogic;
import be.iminds.ilabt.jfed.espec.model.*;
import be.iminds.ilabt.jfed.espec.parser.ESpecConstants;
import be.iminds.ilabt.jfed.espec.parser.ESpecDefaults;
import be.iminds.ilabt.jfed.espec.util.ChModPermissionsUtil;
import be.iminds.ilabt.jfed.espec.util.ESpecLogger;
import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.highlevel.jobs.*;
import be.iminds.ilabt.jfed.highlevel.util.ExternalFileUtil;
import be.iminds.ilabt.jfed.highlevel.util.StringMemorySourceFile;
import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnection;
import be.iminds.ilabt.jfed.rspec.model.BasicStringRspec;
import be.iminds.ilabt.jfed.rspec.model.ModelRspec;
import be.iminds.ilabt.jfed.rspec.model.RspecNode;
import be.iminds.ilabt.jfed.util.AnsibleFileWriter;
import javafx.application.Platform;
import javafx.beans.property.SimpleStringProperty;
import javafx.beans.property.StringProperty;
import net.schmizz.sshj.SSHClient;
import net.schmizz.sshj.connection.ConnectionException;
import net.schmizz.sshj.connection.channel.direct.Session;
import net.schmizz.sshj.sftp.FileAttributes;
import net.schmizz.sshj.sftp.SFTPClient;
import net.schmizz.sshj.transport.TransportException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.event.Level;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.*;
import java.security.KeyPair;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.util.*;

/**
 * User: twalcari
 * Date: 5/27/2016
 * Time: 11:51
 */
public class ExperimentSpecificationServicesAnsibleState extends SlicedState<ExperimentSpecificationServicesAnsibleState.ExperimentSpecificationServicesAnsibleStateSlice> {
    private static final Logger LOG = LoggerFactory.getLogger(ExperimentSpecificationServicesAnsibleState.class);

    @Nonnull private final SetupSoftwareExperimentJob job;
    @Nonnull private final Experiment experiment;
    @Nonnull private final ExperimentSpecification experimentSpecification;
    @Nonnull private final GeniUserProvider geniUserProvider;
    @Nonnull private final ExperimentSpecificationServicesDirsAndUploadsState parent;
    @Nonnull private final AnsibleSpec ansibleSpec;

    @Nonnull private final ESpecLogger eSpecLogger;

    private final boolean useRemoteAnsibleNode;
    @Nonnull private final String ansibleNodeClientId;
    @Nonnull private final RspecNode ansibleNode;


    @Nonnull private final List<ExperimentSpecificationServicesAnsibleStateSlice> slices;

    protected ExperimentSpecificationServicesAnsibleState(@Nonnull SetupSoftwareExperimentJob job,
                                                          @Nonnull GeniUserProvider geniUserProvider,
                                                          @Nonnull ExperimentSpecificationServicesDirsAndUploadsState parent,
                                                          @Nonnull AnsibleSpec ansibleSpec) {
        super("Execute ansible services");
        this.job = job;
        this.geniUserProvider = geniUserProvider;
        this.parent = parent;
        this.ansibleSpec = ansibleSpec;
        this.experiment = job.getExperiment();
        if (job.getExperiment().getExperimentSpecification() == null)
            throw new IllegalStateException("job.getExperiment().getExperimentSpecification() is null");
        this.experimentSpecification = job.getExperiment().getExperimentSpecification();
        this.eSpecLogger = experiment.getExperimentSpecificationLogger();
        experiment.requireSlice();

        /////////////

        assert ESpecLogic.hasAnsibleNode(experimentSpecification);
        this.useRemoteAnsibleNode = ansibleSpec.getAnsibleHostSpec().getHostType() != AnsibleHostSpec.HostType.LOCAL;

        if (!useRemoteAnsibleNode) {
            throw new RuntimeException("LOCAL ansible execute is not yet implemented.");
        } else {
            if (ansibleSpec.getAnsibleHostSpec().getNodeName() == null)
                throw new IllegalStateException("ansible node name should be set");
        }

        assert experiment.getNewRequestRspecSource() != null;
        ModelRspec request = experiment.getNewRequestRspecSource() == null ?
                null :
                experiment.getNewRequestRspecSource().getModelRspec();

        if (request == null) {
            String err = "Cannot find ansible node because parsing request rspec failed";
            LOG.error(err);
            throw new RuntimeException(err);
        }

        RspecNode ansibleNode = request.getNodeByClientId(ansibleSpec.getAnsibleHostSpec().getNodeName());
        if (ansibleNode == null || ansibleNode.getClientId() == null) {
            String err = "Cannot find ansible node \""+ansibleSpec.getAnsibleHostSpec().getNodeName()+"\" in request rspec";
            LOG.error(err);
            throw new RuntimeException(err);
        }
        this.ansibleNode = ansibleNode;
        this.ansibleNodeClientId = ansibleNode.getClientId();

        this.slices = Collections.singletonList(
                new ExperimentSpecificationServicesAnsibleStateSlice(job, ansibleNode, ansibleSpec)
        );
    }

    @Override
    public Collection<ExperimentSpecificationServicesAnsibleStateSlice> getSlices() {
        return slices;
    }

    public class ExperimentSpecificationServicesAnsibleStateSlice extends StateSlice {
        @Nonnull  protected final SetupSoftwareExperimentJob setupSoftwareExperimentJob;
        @Nonnull protected final RspecNode ansibleNode;
        @Nonnull protected final AnsibleSpec ansibleSpec;

        @Nonnull private String galaxyCommand;
        @Nonnull private String playbookCommand;

        //only initialised when there is an SSH connection (relies on homedir location)
        private String ansibleDirPath;

        public ExperimentSpecificationServicesAnsibleStateSlice(@Nonnull SetupSoftwareExperimentJob job,
                                                                @Nonnull RspecNode ansibleNode,
                                                                @Nonnull AnsibleSpec ansibleSpec) {
            super(job);
            this.setupSoftwareExperimentJob = job;
            this.ansibleNode = ansibleNode;
            this.ansibleSpec = ansibleSpec;

            this.galaxyCommand = ansibleSpec.getAnsibleHostSpec().getGalaxyCommand() == null ?
                    ESpecDefaults.ANSIBLE_GALAXY_COMMAND :
                    ansibleSpec.getAnsibleHostSpec().getGalaxyCommand();
            this.playbookCommand = ansibleSpec.getAnsibleHostSpec().getPlaybookCommand() == null ?
                    ESpecDefaults.ANSIBLE_PLAYBOOK_COMMAND :
                    ansibleSpec.getAnsibleHostSpec().getPlaybookCommand();
        }

        @Override
        public String getName() {
            return "Ansible on " + ansibleNode.getClientId();
        }

        @Override
        public ExperimentTaskStatus statefulRun() throws JFedException, InterruptedException {
            SSHClient ssh;
            try {
                ssh = setupSoftwareExperimentJob.getSSHClient(ansibleNode, this);
            } catch (SetupSoftwareExperimentJob.SshException e) {
                LOG.warn("Aborting ExecuteAnsibleServiceStateSlice.statefulRun due to SSH comnection problem", e);
                return ExperimentTaskStatus.FAILED;
            }
            assert ssh != null;
            ExternalFileUtil externalFileUtil = new ExternalFileUtil(ssh);

            String ansibleDirContentPath = parent.getRemotePath(ESpecConstants.DirContent.ANSIBLE, ansibleNodeClientId);
            if (ansibleDirContentPath == null)
                throw new IllegalStateException("Unknown ansible content dir");
            this.ansibleDirPath = ensureEndsWithSlash(ansibleDirContentPath);

            eSpecLogger.firePreAnsible();

            PrepareAnsibleDirState prepareAnsibleDirState = new PrepareAnsibleDirState(ssh);
            setAndRunState(prepareAnsibleDirState);

            ExperimentTaskStatus worstStatus = prepareAnsibleDirState.getStatus();

            for (AnsibleGalaxySpec gs : ansibleSpec.getAnsibleGalaxySpecs()) {
                RunGalaxyState executeGalaxyState = new RunGalaxyState(gs, ssh);
                setAndRunState(executeGalaxyState);
                worstStatus = ExperimentTaskStatus.worstOf(worstStatus, executeGalaxyState.getStatus());
            }

            for (AnsiblePlaybookSpec pbs : ansibleSpec.getAnsiblePlaybookSpecs()) {
                ExecutePlayBookState executePlayBookState = new ExecutePlayBookState(pbs, ssh);
                setAndRunState(executePlayBookState);
                worstStatus = ExperimentTaskStatus.worstOf(worstStatus, executePlayBookState.getStatus());
            }

            setupSoftwareExperimentJob.closeSSH(ansibleNode);

            eSpecLogger.firePostAnsibleAll(worstStatus == ExperimentTaskStatus.SUCCESS);

            return worstStatus;
        }

        public class PrepareAnsibleDirState extends State {
            @Nonnull private final SSHClient ssh;

            protected PrepareAnsibleDirState(@Nonnull SSHClient ssh) {
                super("Preparing ansible dir '"+ansibleDirPath+"'");

                this.ssh = ssh;

                if (experiment.getSlice().getManifestRspec() == null)
                    throw new IllegalStateException("Cannot prepare ansible node because there is no manifest RSpec.");
                String manifestString = experiment.getSlice().getManifestRspec().getRspecXmlString();
                if (manifestString == null || manifestString.isEmpty())
                    throw new IllegalStateException("Cannot prepare ansible node because there is an empty manifest RSpec.");
            }

            @Nonnull
            @Override
            protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
                KeyPair ansibleKeyPair = experiment.getKeypairs().values().iterator().next();

                if (experiment.getSlice().getManifestRspec() == null)
                    throw new IllegalStateException("Cannot prepare ansible node because there is no manifest RSpec.");
                String manifestString = experiment.getSlice().getManifestRspec().getRspecXmlString();
                if (manifestString == null || manifestString.isEmpty())
                    throw new IllegalStateException("Cannot prepare ansible node because there is an empty manifest RSpec.");

                AnsibleFileWriter ansibleFileWriter = AnsibleFileWriter.createWithCopiedPrivateKey(
                        new BasicStringRspec(manifestString),
                        ansibleKeyPair.getPrivate(),
                        ansibleKeyPair.getPublic(),
                        geniUserProvider.getLoggedInGeniUser(),
                        null /* this is meant for testbed node, so it never needs any proxy */,
                        geniUserProvider.getLoggedInGeniUser().getUserUrn().getResourceName());
                ansibleFileWriter.addExcludedClientId(ansibleNodeClientId);

                List<AnsibleFileWriter.AnsibleFileInfo> ansibleFiles = ansibleFileWriter.getFilesToWrite(false);

                try (SFTPClient sftpClient = ssh.newSFTPClient()) {
                    for (AnsibleFileWriter.AnsibleFileInfo ansibleFileInfo : ansibleFiles) {
                        String fullPath = ansibleDirPath + "" + ansibleFileInfo.getFilename();
                        try {
                            ssh.newSCPFileTransfer().upload(
                                    new StringMemorySourceFile(ansibleFileInfo.getFilename(), ansibleFileInfo.getContent()),
                                    fullPath);

                            if (ansibleFileInfo.isPrivateFile()) {
                                //NOTE: sftpClient.chmod and other permsision related commands/parameters in sshj SFTP do not work reliably...
                                sftpClient.chmod(fullPath, ChModPermissionsUtil.toMask("u=r"));

                                //So we also try another way as backup
//                            final FileAttributes attrs = sftpClient.statExistence(fullPath);
//                            if (attrs.getPermissions() ?) {
                                ssh.startSession().exec("chmod u=r " + fullPath);
                            }
                        } catch (IOException ex) {
                            LOG.error("Error while uploading inventory-file to {} on {}", fullPath, ansibleNodeClientId, ex);
                            updateMessage(String.format("Error while uploading inventory-file to %s on %s: %s", fullPath, ansibleNodeClientId, ex.getMessage()));
                            return ExperimentTaskStatus.FAILED;
                        }
                    }
                } catch (IOException ex) {
                    LOG.error("Error while initialising SFTP client for ansible preparations", ex);
                    updateMessage(String.format("Error while initialising SFTP client for ansible preparations: %s", ex.getMessage()));
                    return ExperimentTaskStatus.FAILED;
                }

                return ExperimentTaskStatus.SUCCESS;
            }
        }

        public class RunGalaxyState extends State {
            @Nonnull private final AnsibleGalaxySpec galaxySpec;
            @Nonnull private final SSHClient ssh;

            @Nonnull private final StringProperty galaxyInputStream, galaxyErrorStream;

            protected RunGalaxyState(@Nonnull AnsibleGalaxySpec galaxySpec, @Nonnull SSHClient ssh) {
                super("Running galaxy requirements '" + galaxySpec.getDesc() + "'");

                this.galaxySpec = galaxySpec;
                this.ssh = ssh;

                this.galaxyInputStream = new SimpleStringProperty("");
                this.galaxyErrorStream = new SimpleStringProperty("");
            }

            @Nonnull
            @Override
            protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
                try {
                    String remotePath = parent.getRemotePath(galaxySpec, ansibleNodeClientId);
                    if (remotePath == null) {
                        String err = "Did not find remote path of ansible galaxy requirements. Bug in ansible step?";
                        LOG.error(err);
                        updateMessage(err);
                        throw new RuntimeException(err);
                        //                    return ExperimentTaskStatus.FAILED;
                    }

                    String logPath;
                    if (galaxySpec.getLog() == null) {
                        if (experimentSpecification.getConfig().isStoreRemoteLogsByDefault())
                            logPath = ESpecDefaults.deriveAutomaticGalaxyLogFile(remotePath);
                        else
                            logPath = null;
                    } else {
                        logPath = parent.getRemoteLogPath(galaxySpec, ansibleNodeClientId);
                    }

                    //isntall ansible galaxy requirements
                    String galaxyCommandString = String.format(
                            "%s install -r %s -p roles", galaxyCommand, remotePath);
                    LOG.debug("Executing galaxy command: "+galaxyCommandString);

                    String command = ExperimentSpecificationServicesSingleExecuteState.createSshCommand(
                            galaxyCommandString,
                            ansibleDirPath,
                            logPath,
                            false,
                            experimentSpecification
                    );

                    LOG.debug("Actual SSH exec command: "+command);

                    Session.Command c =
                            ssh.startSession().exec(command);

                    new InputStreamToStringPropertyReaderThread(c.getInputStream(), galaxyInputStream, Level.INFO, "GALAXY-OUTPUT: ").start();
                    new InputStreamToStringPropertyReaderThread(c.getErrorStream(), galaxyErrorStream, Level.WARN, "GALAXY-ERROR: ").start();

                    c.join();

                    if (c.getExitStatus() != 0) {
                        updateMessage("ansible-galaxy execution was unsuccessful. Aborting");
                        return ExperimentTaskStatus.FAILED;
                    }
                } catch (TransportException | ConnectionException ex) {
                    throw new JFedException("SSH-exception while executing ansible: " + ex.getMessage(), ex);
                }

                return ExperimentTaskStatus.SUCCESS;
            }

            @Nonnull
            public StringProperty getGalaxyInputStream() {
                return galaxyInputStream;
            }

            @Nonnull
            public StringProperty getGalaxyErrorStream() {
                return galaxyErrorStream;
            }
        }

        public class ExecutePlayBookState extends State {
            @Nonnull private final AnsiblePlaybookSpec playbook;
            @Nonnull private final SSHClient ssh;

            private final StringProperty playbookInputStream, playbookErrorStream;


            protected ExecutePlayBookState(@Nonnull AnsiblePlaybookSpec playbook, @Nonnull SSHClient ssh) {
                super("Executing ansible-playbook '" + playbook.getDesc() + "'");

                this.playbook = playbook;
                this.ssh = ssh;

                this.playbookInputStream = new SimpleStringProperty("");
                this.playbookErrorStream = new SimpleStringProperty("");
            }

            @Nonnull
            @Override
            protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
                String remotePath = parent.getRemotePath(playbook, ansibleNodeClientId);
                if (remotePath == null) {
                    eSpecLogger.firePostAnsiblePlaybook(playbook, "not found", ansibleNodeClientId, false);
                    String err = "Did not find remote path of ansible playbook. Bug in ansible step?";
                    LOG.error(err);
                    updateMessage(err);
                    throw new RuntimeException(err);
                    //                    return ExperimentTaskStatus.FAILED;
                }

                String logPath;
                if (playbook.getLog() == null) {
                    if (experimentSpecification.getConfig().isStoreRemoteLogsByDefault())
                        logPath = ESpecDefaults.deriveAutomaticPlaybookLogFile(remotePath);
                    else
                        logPath = null;
                } else {
                    logPath = parent.getRemoteLogPath(playbook, ansibleNodeClientId);
                }

                try {
                    String arguments = "";
                    arguments += remotePath;

                    //ansible debug uses one to 3 v's  -v  -vv  or  -vvv
                    if (playbook.isDebug()) {
                        arguments += " -";
                        for (int i = 0; i < playbook.getDebug(); i++)
                            arguments += "v";
                    }

                    if (playbook.hasExtraVars()) {
                        arguments += "--extra-vars '"+playbook.getExtraVarsString()+"'";
                    }

                    //don't need this if AnsibleFileWriter writes the ansible files and uses ansible config
//                    arguments += " -i '"+ansibleInventoryPath+"'";

                    //don't need this if AnsibleFileWriter writes the ansible files
//                    arguments += " -e 'host_key_checking=False'";

                    String playbookCommandString = playbookCommand + " " + arguments;

                    String command = ExperimentSpecificationServicesSingleExecuteState.createSshCommand(
                            playbookCommandString,
                            ansibleDirPath,
                            logPath,
                            false,
                            experimentSpecification
                    );

                    LOG.debug("Executing playbook command: "+playbookCommandString);
                    LOG.debug("Actual SSH exec command: "+command);
                    Session.Command c =
                            ssh.startSession().exec(command);

                    new InputStreamToStringPropertyReaderThread(c.getInputStream(), playbookInputStream, Level.INFO, "PLAYBOOK-OUTPUT: ").start();
                    new InputStreamToStringPropertyReaderThread(c.getErrorStream(), playbookErrorStream, Level.WARN, "PLAYBOOK-ERROR: ").start();

                    c.join();

                    if (c.getExitStatus() != 0) {
                        updateMessage("ansible-playbook execution was unsuccessful.");
                        eSpecLogger.firePostAnsiblePlaybook(playbook, remotePath, ansibleNodeClientId, false);
                        return ExperimentTaskStatus.FAILED;
                    } else {
                        eSpecLogger.firePostAnsiblePlaybook(playbook, remotePath, ansibleNodeClientId, true);
                        return ExperimentTaskStatus.SUCCESS;
                    }
                } catch (TransportException | ConnectionException ex) {
                    eSpecLogger.firePostAnsiblePlaybook(playbook, remotePath, ansibleNodeClientId, false);
                    throw new JFedException("SSH-exception while executing ansible: " + ex.getMessage(), ex);
                }
            }

            public StringProperty getPlaybookInputStream() {
                return playbookInputStream;
            }

            public StringProperty getPlaybookErrorStream() {
                return playbookErrorStream;
            }
        }
    }

    private static class InputStreamToStringPropertyReaderThread extends Thread {

        private final InputStream inputStream;
        private final StringProperty output;

        private final String logPrefix;
        private final LogFunction logFunction;


        private InputStreamToStringPropertyReaderThread(InputStream inputStream, StringProperty output,
                                                        @Nullable Level logLevel, String logPrefix) {
            this.inputStream = inputStream;
            this.output = output;
            this.logPrefix = logPrefix;

            if (logLevel != null) {
                switch (logLevel) {
                    case TRACE:
                        logFunction = LOG::trace;
                        break;
                    case DEBUG:
                        logFunction = LOG::debug;
                        break;
                    case INFO:
                        logFunction = LOG::info;
                        break;
                    case WARN:
                        logFunction = LOG::warn;
                        break;
                    case ERROR:
                        logFunction = LOG::error;
                        break;
                    default:
                        logFunction = null;
                }

            } else
                logFunction = null;
        }

        @Override
        public void run() {

            try {
                BufferedReader stdInput = new BufferedReader(new InputStreamReader(inputStream));

                String s;
                while ((s = stdInput.readLine()) != null) {
                    System.out.println(s);
                    String str = s;

                    if (logFunction != null) {
                        logFunction.log(logPrefix + str);
                    }

                    Platform.runLater(() -> output.setValue(output.get() + "\n" + str));

                }

                inputStream.close();
            } catch (IOException ex) {
                LOG.error("Error while reading from InputStream to StringProperty", ex);
            }

        }

        private interface LogFunction {
            void log(String text);
        }
    }

    @Nonnull
    private static String ensureEndsWithSlash(@Nonnull String p) {
        if (p.endsWith("/")) return p;
        return p + "/";
    }

    @Nonnull
    private static String ensureStartsWithSlash(@Nonnull String p) {
        if (p.startsWith("/")) return p;
        return "/" + p;
    }
}
